"""
Tests of api/views.py.
"""

from unittest.mock import patch

from ddt import ddt
from django.urls import reverse
from opaque_keys.edx.keys import CourseKey
from requests.exceptions import Timeout
from rest_framework.permissions import IsAuthenticated
from rest_framework.status import (
    HTTP_200_OK,
    HTTP_201_CREATED,
    HTTP_400_BAD_REQUEST,
    HTTP_403_FORBIDDEN,
    HTTP_504_GATEWAY_TIMEOUT,
)
from rest_framework.test import APITestCase

from translatable_xblocks.api.views import AiTranslationService


@ddt
class TestConfigApiGet(APITestCase):
    """Tests for ConfigApi GET endpoint."""

    valid_course_key = "course-v1:edX+DemoX+Demo_Course"
    configured_languages = [
        {"code": "en", "label": "English"},
        {"code": "gf", "label": "Gallifreyan"},
    ]

    def get_config_request(self, course_id):
        """Wraps the client call for ease of use."""
        return self.client.get(
            reverse("translatable-xblocks-config"), {"course_id": course_id}
        )

    @patch.object(AiTranslationService, "get_course_translation_languages")
    @patch("translatable_xblocks.api.views.Config")
    @patch("translatable_xblocks.api.views.get_transcript_languages_val")
    def test_get_config(
        self,
        mock_get_transcript_languages,
        mock_config,
        mock_get_course_translation_languages,
    ):
        with patch.object(IsAuthenticated, "has_permission", return_value=True):
            # Given a valid course & settings
            course_id = self.valid_course_key

            mock_available = "foo"
            mock_enabled = "bar"
            mock_get_translations_language = self.configured_languages

            mock_config.xpert_translations_available_for_course.return_value = (
                mock_available
            )
            mock_config.xpert_translations_enabled_for_course.return_value = (
                mock_enabled
            )
            mock_get_course_translation_languages.return_value = (
                mock_get_translations_language
            )
            mock_get_transcript_languages.return_value = ["en", "gf"]

            # When I ask for the config
            response = self.get_config_request(course_id)

            # Then the response returns successfully...
            self.assertEqual(response.status_code, HTTP_200_OK)

            # ... with the expected data
            self.assertEqual(
                response.data["available_translation_languages"],
                self.configured_languages,
            )
            self.assertEqual(response.data["feature_available"], mock_available)
            self.assertEqual(response.data["feature_enabled"], mock_enabled)
            self.assertEqual(response.data["transcription_languages"], ["en", "gf"])

    @patch.object(AiTranslationService, "get_course_translation_languages")
    @patch("translatable_xblocks.api.views.Config")
    def test_get_config_timeout(
        self,
        mock_config,
        mock_get_course_translation_languages,
    ):
        with patch.object(IsAuthenticated, "has_permission", return_value=True):
            # Given a valid course & settings
            course_id = self.valid_course_key

            mock_available = "foo"
            mock_enabled = "bar"

            mock_config.xpert_translations_available_for_course.return_value = (
                mock_available
            )
            mock_config.xpert_translations_enabled_for_course.return_value = (
                mock_enabled
            )

            # ... with a timeout
            mock_get_course_translation_languages.side_effect = Timeout

            # When I ask for the config
            response = self.get_config_request(course_id)

            # Then the response returns a 504 error
            self.assertEqual(response.status_code, HTTP_504_GATEWAY_TIMEOUT)

    def test_get_config_missing_course_id(self):
        with patch.object(IsAuthenticated, "has_permission", return_value=True):
            # Given a missing course_id
            # When I ask for the config
            response = self.client.get(reverse("translatable-xblocks-config"))

            # Then I get a 400 error
            self.assertEqual(response.status_code, HTTP_400_BAD_REQUEST)

    def test_get_config_bad_course_id(self):
        with patch.object(IsAuthenticated, "has_permission", return_value=True):
            # Given a missing course_id
            course_id = "blarg"

            # When I ask for the config
            response = self.get_config_request(course_id)

            # Then I get a 400 error
            self.assertEqual(response.status_code, HTTP_400_BAD_REQUEST)

    def test_get_config_unauthenticated(self):
        # Given I'm not authenticated
        with patch.object(IsAuthenticated, "has_permission", return_value=False):
            course_id = self.valid_course_key

            # When I ask for the config
            response = self.get_config_request(course_id)

            # Then I get a 403 error
            self.assertEqual(response.status_code, HTTP_403_FORBIDDEN)


@ddt
class TestConfigApiSet(APITestCase):
    """Tests for ConfigApi POST endpoint."""

    valid_course_key = "course-v1:edX+DemoX+Demo_Course"
    configured_languages = [
        {"code": "en", "label": "English"},
        {"code": "gf", "label": "Gallifreyan"},
    ]

    def set_config_request(
        self, course_id, feature_enabled, available_translation_languages
    ):
        """Wraps the client call for ease of use."""
        request_data = {
            "course_id": course_id,
            "feature_enabled": feature_enabled,
            "available_translation_languages": available_translation_languages,
        }
        return self.client.post(
            reverse("translatable-xblocks-config"), request_data, format="json"
        )

    @patch.object(AiTranslationService, "post_course_translation_languages")
    @patch("translatable_xblocks.api.views.get_user_role")
    @patch("translatable_xblocks.api.views.Config.enable_xpert_translations_for_course")
    @patch("translatable_xblocks.api.views.Config")
    def test_set_config(
        self,
        mock_config,
        mock_set_feature_enabled,
        mock_get_user_role,
        mock_post_course_translation_languages,
    ):
        with patch.object(IsAuthenticated, "has_permission", return_value=True):
            # Given a valid course & settings
            course_id = self.valid_course_key

            mock_config.xpert_translations_available_for_course.return_value = "foo"
            mock_config.xpert_translations_enabled_for_course.return_value = "bar"
            mock_post_translations_language = {
                "course_id": course_id,
                "available_unit_translation_languages": self.configured_languages,
            }

            mock_post_course_translation_languages.return_value = (
                mock_post_translations_language
            )

            # When I, as staff, try to set the config
            mock_get_user_role.return_value = "staff"
            new_setting_value = "baz"

            response = self.set_config_request(
                course_id, new_setting_value, self.configured_languages
            )

            # Then the response returns successfully...
            self.assertEqual(response.status_code, HTTP_201_CREATED)

            # ... calling the right utility function
            mock_set_feature_enabled.assert_called_with(
                CourseKey.from_string(course_id), new_setting_value
            )

            mock_post_course_translation_languages.assert_called_with(
                mock_post_translations_language
            )

    @patch.object(AiTranslationService, "get_course_translation_languages")
    @patch("translatable_xblocks.api.views.get_user_role")
    @patch("translatable_xblocks.api.views.Config.enable_xpert_translations_for_course")
    @patch("translatable_xblocks.api.views.Config")
    def test_get_config_missing_available_translation_languages(
        self,
        mock_config,
        mock_set_feature_enabled,
        mock_get_user_role,
        mock_get_course_translation_languages,
    ):
        course_id = self.valid_course_key

        mock_config.xpert_translations_available_for_course.return_value = "foo"
        mock_config.xpert_translations_enabled_for_course.return_value = "bar"
        mock_post_translations_language = {
            "course_id": course_id,
            "available_unit_translation_languages": self.configured_languages,
        }

        mock_get_course_translation_languages.return_value = (
            mock_post_translations_language
        )

        # When I, as staff, try to set the config
        mock_get_user_role.return_value = "staff"
        new_setting_value = "baz"
        with patch.object(IsAuthenticated, "has_permission", return_value=True):
            # Given a missing enabled flag
            course_id = self.valid_course_key

            # When I ask for the config
            response = self.client.post(
                reverse("translatable-xblocks-config"),
                {"course_id": course_id, "feature_enabled": new_setting_value},
            )

            # Then the response returns successfully...
            self.assertEqual(response.status_code, HTTP_201_CREATED)

            # mock_set_feature_enabled.assert_called_once()
            mock_set_feature_enabled.assert_called_with(
                CourseKey.from_string(course_id), new_setting_value
            )

            mock_get_course_translation_languages.assert_called_with(course_id)

    def test_get_config_missing_course_id(self):
        with patch.object(IsAuthenticated, "has_permission", return_value=True):
            # Given a missing course_id
            # When I ask for the config
            response = self.client.post(
                reverse("translatable-xblocks-config"), {"feature_enabled": True}
            )

            # Then I get a 400 error
            self.assertEqual(response.status_code, HTTP_400_BAD_REQUEST)

    def test_get_config_bad_course_id(self):
        with patch.object(IsAuthenticated, "has_permission", return_value=True):
            # Given a missing course_id
            course_id = "blarg"

            # When I ask for the config
            response = self.set_config_request(course_id, True, {})

            # Then I get a 400 error
            self.assertEqual(response.status_code, HTTP_400_BAD_REQUEST)

    def test_get_config_missing_feature_enabled(self):
        with patch.object(IsAuthenticated, "has_permission", return_value=True):
            # Given a missing enabled flag
            course_id = self.valid_course_key

            # When I ask for the config
            response = self.client.post(
                reverse("translatable-xblocks-config"), {"course_id": course_id}
            )

            # Then I get a 400 error
            self.assertEqual(response.status_code, HTTP_400_BAD_REQUEST)

    def test_get_config_unauthenticated(self):
        # Given I'm not authenticated
        with patch.object(IsAuthenticated, "has_permission", return_value=False):
            course_id = self.valid_course_key

            # When I ask for the config
            response = self.set_config_request(course_id, True, {})

            # Then I get a 403 error
            self.assertEqual(response.status_code, HTTP_403_FORBIDDEN)



"""Config specific to the production environment"""


def plugin_settings(settings):
    """
    App-specific settings
    """

    settings.AI_TRANSLATIONS_OAUTH_APP_NAME = "ai-translations-backend-service"



"""
Translatable version of edx-platform/xmodule.html_block
"""

# pylint:  disable=unnecessary-lambda-assignment

from xblock.core import XBlock
from xblock.fields import Scope
from xmodule.html_block import HtmlBlock as OverriddenHtmlBlock

from translatable_xblocks.base import TranslatableXBlock
from translatable_xblocks.fields import TranslatableHTML

# Make '_' a no-op so we can scrape strings. Using lambda instead of
#  `django.utils.translation.gettext_noop` because Django cannot be imported in this file
_ = lambda text: text


@XBlock.needs("i18n")
@XBlock.needs("mako")
@XBlock.needs("user")
class HtmlBlock(TranslatableXBlock, OverriddenHtmlBlock):
    """
    Translatable version of the HtmlBlock.

    Here, we only have to translate the data field.
    """

    data = TranslatableHTML(
        help=_("Html contents to display for this block"),
        default="",
        scope=Scope.content,
    )

    @XBlock.supports("multi_device")
    def student_view(self, context, **kwargs):
        return super().student_view(context, **kwargs)



"""
Utilities, split out for easier testing & mocking.
"""


def get_course_from_modulestore(course_id):
    """
    Get course from modulestore.
    For whatever reason, this has to be a runtime import or it breaks a bunch
    of things.
    """
    # pylint: disable=import-outside-toplevel
    from xmodule.modulestore.django import modulestore

    return modulestore().get_course(course_id)



"""API v1 URLs."""

from django.urls import path

from translatable_xblocks.api.views import ConfigAPI

urlpatterns = [
    path(
        "config/",
        ConfigAPI.as_view(),
        name="translatable-xblocks-config",
    ),
]



"""
Tests of waffle.py.
"""

from unittest.mock import patch

import ddt
from django.test import TestCase

from translatable_xblocks.config import (
    WHOLE_COURSE_TRANSLATIONS_ENABLED_FLAG,
    TranslatableXBlocksFeatureConfig,
)


class MockCourseWaffle:
    """Mock CourseWaffleFlag behavior."""

    def __init__(self, enabled):
        self.enabled = enabled

    def is_enabled(self, course_key=None):  # pylint: disable=unused-argument
        return self.enabled


@ddt.ddt
class TestConfig(TestCase):
    """Tests for feature configuration / enable / disable."""

    @ddt.data(True, False)
    @patch("translatable_xblocks.config.course_waffle_flag")
    def test_whole_course_translations_available_for_course(
        self, expected_enabled, mock_course_waffle_flag
    ):
        # Given the feature is / not available for a course
        mock_course_waffle_flag.return_value = MockCourseWaffle(expected_enabled)

        # When I query for the feature being available
        actual_enabled = (
            TranslatableXBlocksFeatureConfig.xpert_translations_available_for_course(
                "some_course"
            )
        )

        # Then I get the correct result
        self.assertEqual(expected_enabled, actual_enabled)

    @ddt.unpack
    @ddt.data(
        [False, False, False],
        [False, True, False],
        [True, False, False],
        [True, True, True],
    )
    @patch("translatable_xblocks.config.course_waffle_flag")
    def test_whole_course_translations_enabled_for_course(
        self,
        available_flag_value,
        enabled_flag_value,
        expected_enabled_result,
        mock_course_waffle_flag,
    ):
        # Given the feature is / not enabled / available
        mock_course_waffle_flag.side_effect = [
            MockCourseWaffle(enabled_flag_value),
            MockCourseWaffle(available_flag_value),
        ]

        # When I query for the feature being enabled
        actual_enabled = (
            TranslatableXBlocksFeatureConfig.xpert_translations_enabled_for_course(
                "some_course"
            )
        )

        # Then I get the correct result
        # (enabled ONLY when both available and enabled are True)
        self.assertEqual(expected_enabled_result, actual_enabled)

    @patch("translatable_xblocks.config.course_waffle_flag")
    @patch("translatable_xblocks.config.import_course_waffle_flag_override")
    def test_set_xpert_translations_enabled_for_course(
        self,
        mock_waffle_flag_override,
        mock_is_available,
    ):
        # Given that the translation feature is available
        mock_is_available.is_enabled.return_value = True
        course_id = "some_course"

        # When I try to set the feature to be enabled
        TranslatableXBlocksFeatureConfig.enable_xpert_translations_for_course(
            course_id, True
        )

        # Then I set the waffle flag course override
        mock_waffle_flag_override().objects.create.assert_called_once_with(
            waffle_flag=WHOLE_COURSE_TRANSLATIONS_ENABLED_FLAG,
            course_id=course_id,
            enabled=True,
            override_choice="on",
        )

    @patch(
        "translatable_xblocks.config.TranslatableXBlocksFeatureConfig.xpert_translations_available_for_course"
    )
    @patch("translatable_xblocks.config.import_course_waffle_flag_override")
    def test_set_xpert_translations_disabled_for_course(
        self,
        mock_waffle_flag_override,
        mock_is_available,
    ):
        # Given that the translation feature is available
        mock_is_available.return_value = True
        course_id = "some_course"

        # When I try to set the feature to be disabled
        TranslatableXBlocksFeatureConfig.enable_xpert_translations_for_course(
            course_id, False
        )

        # Then I set the waffle flag course override
        mock_waffle_flag_override().objects.create.assert_called_once_with(
            waffle_flag=WHOLE_COURSE_TRANSLATIONS_ENABLED_FLAG,
            course_id=course_id,
            enabled=True,
            override_choice="off",
        )

    @patch(
        "translatable_xblocks.config.TranslatableXBlocksFeatureConfig.xpert_translations_available_for_course"
    )
    def test_set_xpert_translations_disallowed(
        self,
        mock_is_available,
    ):
        # Given that the translation feature is NOT available
        mock_is_available.return_value = False
        course_id = "some_course"

        # When I try to set the feature to be disabled
        # Then I raise an error
        with self.assertRaises(ValueError):
            TranslatableXBlocksFeatureConfig.enable_xpert_translations_for_course(
                course_id, False
            )



"""Common config across environments"""


def plugin_settings(settings):
    """
    App-specific settings
    """

    # TTL in seconds for successful request to store in cache (1 day)
    settings.REQUEST_CACHE_SUCCESS_TIMEOUT = 86400

    # TTL in seconds for successful config request to store in cache (15 minutes)
    settings.CONFIG_REQUEST_CACHE_TTL = 900

    # TIL in seconds for failure request to store in cache (5 minutes)
    settings.REQUEST_CACHE_FAILURE_TIMEOUT = 300

    # Timeout in seconds for API requests
    settings.AI_TRANSLATIONS_API_TIMEOUT = 15

    # Provider type for AI transcripts
    # See ref: https://github.com/openedx/edx-val/blob/b7c5db6c95c9a8655e1f3d58429c2e52b4ead66e/edxval/models.py#L381
    settings.AI_TRANSCRIPTS_PROVIDER_TYPE = "edx_ai_translations"



"""
Services for AI Translations.
"""

import json
import logging
from datetime import datetime
from hashlib import sha256

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from edx_django_utils.cache.utils import TieredCache
from edx_rest_api_client.client import OAuthAPIClient
from oauth2_provider.models import Application
from rest_framework import status

from translatable_xblocks.utils import (
    reinsert_base64_images,
    replace_img_base64_with_placeholder,
)

logger = logging.getLogger(__name__)


class AiTranslationService:
    """
    A service which communicates with ai-translations for translation-related tasks.
    """

    def __init__(self):
        """Initialize translations service."""
        self._client = None

    @property
    def client(self):
        """Create client for communicating with ai-translations, singleton."""
        if not self._client:
            self._client = self._init_translations_client()
        return self._client

    def _init_translations_client(self):
        """Initialize OAuth connection to ai-translations."""
        application_name = settings.AI_TRANSLATIONS_OAUTH_APP_NAME

        try:
            oauth_client = Application.objects.get(name=application_name)
            return OAuthAPIClient(
                settings.LMS_ROOT_URL,
                oauth_client.client_id,
                oauth_client.client_secret,
            )
        except ObjectDoesNotExist:
            logger.error(f"OAuth Application {application_name} not found.")
            return None

    @staticmethod
    def get_content_hash(content):
        """Generate hash of provided content, using sha256."""
        return sha256(content.encode("utf-8")).hexdigest()

    @staticmethod
    def translation_cache_key(source_language, target_language, content_hash):
        """Generate cache key, using get_content_hash result for content hash."""
        return f"ai_translation.{source_language}.{target_language}.{content_hash}"

    @staticmethod
    def config_cache_key(course_id):
        """Generate cache key for course configuration."""
        return f"ai_translation.config.{course_id}"

    # pylint: disable=too-many-positional-arguments
    def translate(
        self,
        content,
        source_language,
        target_language,
        block_id,
        user_id,
        mimetype="text/html",
        cache_override=None,
    ):
        """
        Request translated version of text / HTML from Translations IDA, checking for cached version first.

        Args:
            content: HTML / text / XML content to transform
            source_language: ISO language code of original content
            target_language: ISO language code to translate content into
            block_id: block ID of originating content (if applicable), used only for logging
            user_id: user ID requesting translation, used only for logging

        Kwargs:
            mimetype (default="text/html"): provided in cases that require different handling (e.g. "application/xml")
            cache_override (default=None): arguments for overriding default cache behavior. Options:
            - "reset" instructs Translations IDA to reset the cache entry for this content.
        """
        content_hash = AiTranslationService.get_content_hash(content)
        block_id_str = str(block_id)

        logger.info(
            f"---- XBlock Translation Attempt ---- "
            f"Timestamp: {datetime.now()} "
            f"Block ID: {block_id_str} "
            f"Hash: {content_hash} "
            f"Source Language: {source_language} "
            f"Target Language: {target_language} "
            f"User ID: {user_id} "
            f"Mimetype: {mimetype} "
            f"Cache Override: {cache_override} "
        )

        # Check for cached response first, clearing if requested
        cache_key = AiTranslationService.translation_cache_key(
            source_language, target_language, content_hash
        )
        if cache_override == "reset":
            logger.info(f"Clearing cache for key: {cache_key}")
            TieredCache.delete_all_tiers(cache_key)
        cache_result = TieredCache.get_cached_response(cache_key)

        # Cached response found, return response
        if cache_result.is_found:
            # A cache value of None (specifically where we expect there to be content)
            # is our signal that there was a cached error. Raise to avoid constantly
            # hitting Translations IDA while it may be encountiering issues.
            if cache_result.value is None and content is not None:
                exception_string = (
                    f"Cached failure response: {block_id_str} "
                    f"for languages {source_language} to {target_language}"
                )
                logger.error(exception_string)
                raise Exception(exception_string)

            # Else, return value
            logger.info(f"Cached Translated Content: Found for key {cache_key}")
            return cache_result.value

        logger.info(
            f"Cached Translated Content: Not Found, requesting translation for key {cache_key}"
        )

        # Cached copy not found, query Translations IDA
        url = f"{settings.AI_TRANSLATIONS_API_URL}/translate-xblock/"
        headers = {
            "content-type": "application/json",
        }

        # sometimes, we have long base64 images in a content
        # this causes the post request not to get to the ai-translations for some unknown reasons
        # this code is added to prevent base64 images from being tranlsated is they aren't required
        processed_content, base64_images = replace_img_base64_with_placeholder(content)

        payload = {
            "block_id": block_id_str,
            "source_language": source_language,
            "target_language": target_language,
            "content": processed_content,
            "content_hash": content_hash,
            "cache_override": cache_override or "none",
        }
        response = self.client.post(url, data=json.dumps(payload), headers=headers)

        if response.status_code != status.HTTP_200_OK:
            TieredCache.set_all_tiers(
                cache_key, None, settings.REQUEST_CACHE_FAILURE_TIMEOUT
            )
            exception_string = (
                f"Failed to get translation for block: {block_id_str} "
                f"for languages {source_language} to {target_language}"
            )
            logger.error(exception_string)
            raise Exception(exception_string)

        translated_content = response.json().get("translated_content")
        if len(base64_images) > 0:
            translated_content = reinsert_base64_images(
                translated_content, base64_images
            )

        # Set cached content
        TieredCache.set_all_tiers(
            cache_key, translated_content, settings.REQUEST_CACHE_SUCCESS_TIMEOUT
        )

        # Return data
        return translated_content

    def get_course_translation_languages(self, course_id):
        """
        Request translation languages available for a course.

        Args:
            course_id: the course_id for the course
        """
        # Check for cached response first
        config_cache_key = AiTranslationService.config_cache_key(course_id)
        cache_result = TieredCache.get_cached_response(config_cache_key)

        # Cached error response, return exception without calling API
        if cache_result.is_found and cache_result.value is None:
            exception_string = (
                f"Cached failure response getting translations config for: {course_id}"
            )
            logger.error(exception_string)
            raise Exception(exception_string)

        # Valid cache response
        if cache_result.is_found:
            return cache_result.value

        # Query AI Translations configuration
        headers = {
            "content-type": "application/json",
        }
        url = f"{settings.AI_TRANSLATIONS_API_URL}/available_languages/"
        response = self.client.get(
            url,
            params={"course_id": course_id},
            headers=headers,
            timeout=settings.AI_TRANSLATIONS_API_TIMEOUT,
        )

        # On failure, cache failure response to avoid duplicate lookups
        if response.status_code != status.HTTP_200_OK:
            TieredCache.set_all_tiers(
                config_cache_key, None, settings.REQUEST_CACHE_FAILURE_TIMEOUT
            )
            exception_string = (
                f"Failed to get translation languages for course: {course_id}"
            )
            logger.error(exception_string)
            raise Exception(exception_string)

        # On success, return the available languages and set cached response
        course_transaltion_languages = response.json().get(
            "available_unit_translation_languages", []
        )
        TieredCache.set_all_tiers(
            config_cache_key,
            course_transaltion_languages,
            settings.CONFIG_REQUEST_CACHE_TTL,
        )
        return course_transaltion_languages

    def post_course_translation_languages(self, data):
        """
        creates/updates translation languages available for a course.

        Args:
            data (dict): Data to send in the post request. Expected keys:
                - course_id: the course_id for the course
                - available_unit_translation_languages: list of languages available for translation
        """
        headers = {
            "content-type": "application/json",
        }
        url = f"{settings.AI_TRANSLATIONS_API_URL}/available_languages/"
        response = self.client.post(url, data=json.dumps(data), headers=headers)
        if response.status_code != status.HTTP_200_OK:
            exception_string = f"Failed to create translation languages for course: {data['course_id']}"
            logger.error(exception_string)
            raise Exception(exception_string)

        # On success, write cached value for the course
        new_course_translation_languages = response.json().get(
            "available_unit_translation_languages", []
        )
        cache_key = AiTranslationService.config_cache_key(data["course_id"])
        TieredCache.set_all_tiers(
            cache_key,
            new_course_translation_languages,
            settings.CONFIG_REQUEST_CACHE_TTL,
        )

        # Return the newly set langauges for the course
        return new_course_translation_languages



"""
Imports from edx-platform.

Our testing environments don't stand up entire edx-platform instances, quite reasonably,
so these imports, while available when we actually run as a plugin, don't exist for
testing. A workaround is to import these at runtime and mock this file for testing.

*Ideally* these are all things that would exist in their own libraries or abstarcted out
but, since that work hasn't been done yet, we're limited to this workaround.
"""

# pylint: disable=import-outside-toplevel


def user_is_admin(request_user):
    """
    Determine whether user is admin.

    Function uses an attribute imported from platform constants, which we could redefine but,
    for consistency, is probably better to import here.
    """
    from common.djangoapps.xblock_django.constants import ATTR_KEY_USER_IS_GLOBAL_STAFF

    return request_user.opt_attrs.get(ATTR_KEY_USER_IS_GLOBAL_STAFF)


def get_user_role(request_user, course_key):
    """
    Get the role of the current user in the course context.

    User roles are not clearly exposed outside of edx-platform, that I'm aware of.
    """
    from lms.djangoapps.courseware.access import get_user_role as platform_get_user_role

    return platform_get_user_role(request_user, course_key)


def get_platform_etree_lib():
    """
    Get platform copy of etree.

    edx-platform has a wrapper around safe_lxml to avoid XML parsing issues.
    This, however, does force us to use the platform versions of lxml and etree.
    """
    from openedx.core.lib.safe_lxml import etree as platform_etree

    return platform_etree


def course_waffle_flag(flag_name, *args, **kwargs):
    """
    Create CourseWaffleFlag model instance.

    edx-platform currently defines its own version of a WaffleFlag, scoped to a course
    to allow overrides on Course/Org level.

    Until this is broken out of platform, however, we have to use this workaround to
    enable tests to work.
    """
    from openedx.core.djangoapps.waffle_utils import CourseWaffleFlag

    # pylint: disable=toggle-missing-annotation
    return CourseWaffleFlag(flag_name, *args, **kwargs)


def import_course_waffle_flag_override():
    """
    Get the WaffleFlagCourseOverrideModel.

    edx-platform currently defines its own version of a WaffleFlag, scoped to a course
    to allow overrides on Course/Org level.

    Until this is broken out of platform, however, we have to use this workaround to
    enable tests to work.
    """
    from openedx.core.djangoapps.waffle_utils.models import (
        WaffleFlagCourseOverrideModel,
    )

    return WaffleFlagCourseOverrideModel


def get_transcript_languages_val(course_id, transcript_provider_type):
    """
    Fetch the transcript languages for a course from edxval.
    """
    from edxval.api import get_transcript_languages

    return get_transcript_languages(course_id, transcript_provider_type)



"""Tests for translatable_xblocks/util.py"""

from unittest import TestCase

from translatable_xblocks.utils import (
    reinsert_base64_images,
    replace_img_base64_with_placeholder,
)


class TestBase64ImageReplacement(TestCase):
    """Tests for replacing base64 images with placeholders."""

    def test_base64_image_replace(self):
        text = (
            "<p>Here is an image:</p>"
            '<img src="data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUA" alt="Example">'
            "<p>Another paragraph.</p>"
            '<img src="data:image/jpg;base64,iBlahBlahBlah" alt="Example2">'
        )

        expected_text = (
            '<p>Here is an image:</p><img src="BASE64_IMG_PLACEHOLDER_0" alt="Example">'
            '<p>Another paragraph.</p><img src="BASE64_IMG_PLACEHOLDER_1" alt="Example2">'
        )

        replaced_text, base64_images = replace_img_base64_with_placeholder(text)

        self.assertEqual(expected_text, replaced_text)
        self.assertEqual(len(base64_images), 2)

        actual_text = reinsert_base64_images(replaced_text, base64_images)
        self.assertEqual(actual_text, text)



"""
Views for the Translatable XBlocks API
"""

import logging

from django.conf import settings
from django.http.response import HttpResponseBadRequest, HttpResponseForbidden
from opaque_keys import InvalidKeyError
from opaque_keys.edx.keys import CourseKey
from requests.exceptions import Timeout
from rest_framework import status
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from translatable_xblocks.ai_translations_service import AiTranslationService
from translatable_xblocks.config import TranslatableXBlocksFeatureConfig as Config
from translatable_xblocks.platform_imports import (
    get_transcript_languages_val,
    get_user_role,
)

logger = logging.getLogger(__name__)


class ConfigAPI(APIView):
    """Get / Set configuration for the translations feature."""

    permission_classes = (IsAuthenticated,)

    ai_translations_service = AiTranslationService()

    def get(self, request):
        """
        List feature configuration
        """

        # Lookup course from request
        course_id = request.GET.get("course_id")
        if not course_id:
            return HttpResponseBadRequest("course_id is required")
        try:
            course_key = CourseKey.from_string(course_id)
        except InvalidKeyError:
            return HttpResponseBadRequest(f"Invalid course course_id: {course_id}")

        # Check that feature is enabled
        feature_available = Config.xpert_translations_available_for_course(course_key)
        feature_enabled = Config.xpert_translations_enabled_for_course(course_key)

        # Get available translation languages
        try:
            available_translation_languages = (
                self.ai_translations_service.get_course_translation_languages(course_id)
            )
        except Timeout:
            exception_string = f"Timeout occurred while getting translation languages for course: {course_id}"
            logger.error(exception_string)
            return Response(
                {"message": exception_string},
                status=status.HTTP_504_GATEWAY_TIMEOUT,
            )

        # Return Data
        return Response(
            {
                "available_translation_languages": available_translation_languages,
                "feature_available": feature_available,
                "feature_enabled": feature_enabled,
                "transcription_languages": get_transcript_languages_val(
                    course_id, settings.AI_TRANSCRIPTS_PROVIDER_TYPE
                ),
            }
        )

    def post(self, request):
        """
        Set feature configuration.
        """

        # Check correct POST body
        if "course_id" not in request.data:
            return HttpResponseBadRequest("course_id is required")
        if "feature_enabled" not in request.data:
            return HttpResponseBadRequest("feature_enabled is required")

        # Pull args from body
        course_id = request.data.get("course_id")
        feature_enabled = request.data.get("feature_enabled")
        available_translation_languages = request.data.get(
            "available_translation_languages"
        )

        # Validate course key
        try:
            course_key = CourseKey.from_string(course_id)
        except InvalidKeyError:
            return HttpResponseBadRequest(f"Invalid course course_id: {course_id}")

        # Check user has instructor / staff permissions
        if get_user_role(request.user, course_key) not in ("staff", "instructor"):
            return HttpResponseForbidden()

        # Set new feature value
        try:
            Config.enable_xpert_translations_for_course(course_key, feature_enabled)

        except ValueError:
            return Response(
                {
                    "available_translation_languages": [],
                    "feature_available": Config.xpert_translations_available_for_course(
                        course_key
                    ),
                    "feature_enabled": Config.xpert_translations_enabled_for_course(
                        course_key
                    ),
                },
                status=status.HTTP_400_BAD_REQUEST,
            )

        # If language config was provided, set those available languages using ai_translations_service
        if available_translation_languages and len(available_translation_languages):
            post_data = {
                "course_id": course_id,
                "available_unit_translation_languages": available_translation_languages,
            }
            available_translation_languages = (
                self.ai_translations_service.post_course_translation_languages(
                    post_data
                )
            )
        # Otherwise, just get existing language config state
        else:
            available_translation_languages = (
                self.ai_translations_service.get_course_translation_languages(course_id)
            )

        return Response(
            {
                "available_translation_languages": available_translation_languages,
                "feature_available": Config.xpert_translations_available_for_course(
                    course_key
                ),
                "feature_enabled": Config.xpert_translations_enabled_for_course(
                    course_key
                ),
            },
            status=status.HTTP_201_CREATED,
        )



"""
translatable_xblocks Django application initialization.
"""

from django.apps import AppConfig
from edx_django_utils.plugins.constants import PluginSettings, PluginURLs


class TranslatableXBlocksConfig(AppConfig):
    """
    Configuration for the translatable_xblocks Django application.
    """

    name = "translatable_xblocks"

    plugin_app = {
        PluginURLs.CONFIG: {
            "lms.djangoapp": {
                PluginURLs.NAMESPACE: "translatable_xblocks",
                PluginURLs.APP_NAME: "translatable_xblocks",
                PluginURLs.REGEX: r"^api/translatable_xblocks/",
                PluginURLs.RELATIVE_PATH: "api.urls",
            },
            "cms.djangoapp": {
                PluginURLs.NAMESPACE: "translatable_xblocks",
                PluginURLs.APP_NAME: "translatable_xblocks",
                PluginURLs.REGEX: r"^api/translatable_xblocks/",
                PluginURLs.RELATIVE_PATH: "api.urls",
            },
        },
        PluginSettings.CONFIG: {
            "lms.djangoapp": {
                "production": {
                    PluginSettings.RELATIVE_PATH: "settings.production",
                },
                "common": {
                    PluginSettings.RELATIVE_PATH: "settings.common",
                },
                "devstack": {
                    PluginSettings.RELATIVE_PATH: "settings.devstack",
                },
            },
            "cms.djangoapp": {
                "production": {
                    PluginSettings.RELATIVE_PATH: "settings.production",
                },
                "common": {
                    PluginSettings.RELATIVE_PATH: "settings.common",
                },
                "devstack": {
                    PluginSettings.RELATIVE_PATH: "settings.devstack",
                },
            },
        },
    }



"""Tests for the AiTranslationService."""

import json
from hashlib import sha256
from unittest import TestCase
from unittest.mock import Mock, PropertyMock, patch

from django.conf import settings
from edx_django_utils.cache.utils import CachedResponse

from translatable_xblocks.ai_translations_service import AiTranslationService


class TestAiTranslationService(TestCase):
    """Tests for AiTranslationService."""

    @patch.object(AiTranslationService, "client", PropertyMock())
    @patch("translatable_xblocks.ai_translations_service.TieredCache")
    def test_translate(self, mock_cache):
        # Given a valid translations service
        ai_service = AiTranslationService()

        # ... and a cache miss for requested data
        mock_cache.get_cached_response = Mock(
            return_value=CachedResponse(False, "foo", "****")
        )
        mock_cache.set_all_tiers = Mock()

        with patch.object(ai_service.client, "post") as mock_post:
            mock_post.return_value.status_code = 200
            mock_post.return_value.json.return_value = {
                "translated_content": "Esto es una prueba"
            }

            content = "This is a test"
            source_language = "en"
            target_language = "es"
            block_id = "block-v1:foo+bar"
            user_id = "1"

            # When I call "translate"
            ai_service.translate(
                content, source_language, target_language, block_id, user_id
            )

            expected_url = "example.com/api/translate-xblock/"
            expected_post_data = {
                "block_id": block_id,
                "source_language": source_language,
                "target_language": target_language,
                "content": content,
                "content_hash": sha256(content.encode("utf-8")).hexdigest(),
                "cache_override": "none",
            }
            expected_headers = {
                "content-type": "application/json",
            }

            # Then I call the endpoint with the correct data
            mock_post.assert_called_with(
                expected_url,
                data=json.dumps(expected_post_data),
                headers=expected_headers,
            )

    @patch.object(AiTranslationService, "client", PropertyMock())
    @patch("translatable_xblocks.ai_translations_service.TieredCache")
    def test_translate_cached(self, mock_cache):
        # Given a valid translations service
        ai_service = AiTranslationService()

        # ... and a cache hit for requested data
        cached_translated_content = "Esto es una prueba"
        mock_cache.get_cached_response = Mock(
            return_value=CachedResponse(True, "foo", cached_translated_content)
        )
        mock_cache.set_all_tiers = Mock()

        with patch.object(ai_service.client, "post") as mock_post:
            content = "This is a test"
            source_language = "en"
            target_language = "es"
            block_id = "block-v1:foo+bar"
            user_id = "1"

            # When I call "translate"
            translated_content = ai_service.translate(
                content, source_language, target_language, block_id, user_id
            )

            # Then the Translations IDA is NOT called
            mock_post.assert_not_called()
            self.assertEqual(translated_content, cached_translated_content)

    @patch.object(AiTranslationService, "client", PropertyMock())
    @patch("translatable_xblocks.ai_translations_service.TieredCache")
    def test_translate_cache_reset(self, mock_cache):
        # Given a valid translations service
        ai_service = AiTranslationService()

        mock_cache.set_all_tiers = Mock()
        mock_cache.delete_all_tiers = Mock()
        mock_cache.get_cached_response = Mock(
            return_value=CachedResponse(False, "foo", "****")
        )

        with patch.object(ai_service.client, "post") as mock_post:
            content = "This is a test"
            content_hash = (
                "c7be1ed902fb8dd4d48997c6452f5d7e509fbcdbe2808b16bcf4edce4c07d14e"
            )
            source_language = "en"
            target_language = "es"
            block_id = "block-v1:foo+bar"
            user_id = "1"

            mock_post.return_value.status_code = 200
            mock_post.return_value.json.return_value = {
                "translated_content": "Esto es una prueba"
            }

            cache_key = AiTranslationService.translation_cache_key(
                source_language, target_language, content_hash
            )

            # When I call "translate" with a request to reset the cache
            ai_service.translate(
                content,
                source_language,
                target_language,
                block_id,
                user_id,
                cache_override="reset",
            )

            # Then we ask the cache to reset
            mock_cache.delete_all_tiers.assert_called_once_with(cache_key)

            # ... and the Translations IDA IS called
            mock_post.assert_called_once()

            # ... and the cache is set again with newly translated content
            mock_cache.set_all_tiers.assert_called_once()

    @patch.object(AiTranslationService, "client", PropertyMock())
    @patch("translatable_xblocks.ai_translations_service.TieredCache")
    def test_translate_error(self, mock_cache):
        # Given a valid translations service
        ai_service = AiTranslationService()

        mock_cache.set_all_tiers = Mock()
        mock_cache.get_cached_response = Mock(
            return_value=CachedResponse(False, "foo", "****")
        )

        with patch.object(ai_service.client, "post") as mock_post:
            content = "This is a test"
            content_hash = (
                "c7be1ed902fb8dd4d48997c6452f5d7e509fbcdbe2808b16bcf4edce4c07d14e"
            )
            source_language = "en"
            target_language = "es"
            block_id = "block-v1:foo+bar"
            user_id = "1"

            # When there is an error with getting a translation
            mock_post.return_value.status_code = 418

            # Then we raise an exception
            with self.assertRaises(Exception):
                ai_service.translate(
                    content,
                    source_language,
                    target_language,
                    block_id,
                    user_id,
                )

            # And cache an error for a short time
            cache_key = AiTranslationService.translation_cache_key(
                source_language, target_language, content_hash
            )
            mock_cache.set_all_tiers.assert_called_once_with(
                cache_key, None, settings.REQUEST_CACHE_FAILURE_TIMEOUT
            )

    @patch.object(AiTranslationService, "client", PropertyMock())
    @patch("translatable_xblocks.ai_translations_service.TieredCache")
    def test_translate_error_cached(self, mock_cache):
        # Given a valid translations service
        ai_service = AiTranslationService()

        mock_cache.set_all_tiers = Mock()

        with patch.object(ai_service.client, "post") as mock_post:
            content = "This is a test"
            content_hash = (
                "c7be1ed902fb8dd4d48997c6452f5d7e509fbcdbe2808b16bcf4edce4c07d14e"
            )
            source_language = "en"
            target_language = "es"
            block_id = "block-v1:foo+bar"
            user_id = "1"

            # When we ask for a translation, but get a cached error response
            cache_key = AiTranslationService.translation_cache_key(
                source_language, target_language, content_hash
            )
            mock_cache.get_cached_response = Mock(
                return_value=CachedResponse(True, cache_key, None)
            )

            # Then we raise an exception
            with self.assertRaises(Exception):
                ai_service.translate(
                    content,
                    source_language,
                    target_language,
                    block_id,
                    user_id,
                )

        # ... and do NOT call the translations IDA
        mock_post.assert_not_called()

    @patch.object(AiTranslationService, "client", PropertyMock())
    @patch("translatable_xblocks.ai_translations_service.TieredCache")
    def test_translate_xml(self, mock_cache):
        # Given a valid translations service
        ai_service = AiTranslationService()

        # ... and a cache miss for requested data
        mock_cache.get_cached_response = Mock(
            return_value=CachedResponse(False, "foo", "****")
        )
        mock_cache.set_all_tiers = Mock()

        with patch.object(ai_service.client, "post") as mock_post:
            # Translated content from Google follows the HTML spec instead of XML...
            # ... causing some transforms that break XML parsing
            translated_html = (
                "<html><body><option correct=False>False</option></body></html>"
            )

            mock_post.return_value.status_code = 200
            mock_post.return_value.json.return_value = {
                "translated_content": translated_html
            }

            content = '<option correct="false">Attributes should be quoted</option>'
            source_language = "en"
            target_language = "es"
            block_id = "block-v1:foo+bar"
            user_id = "1"

            # When I call "translate" for XML by specifying an XML mimetype
            ai_service.translate(
                content,
                source_language,
                target_language,
                block_id,
                user_id,
                mimetype="application/xml",
            )

            expected_url = "example.com/api/translate-xblock/"
            expected_post_data = {
                "block_id": block_id,
                "source_language": source_language,
                "target_language": target_language,
                "content": content,
                "content_hash": sha256(content.encode("utf-8")).hexdigest(),
                "cache_override": "none",
            }
            expected_headers = {
                "content-type": "application/json",
            }

            # Then I call the endpoint with the correct data
            mock_post.assert_called_with(
                expected_url,
                data=json.dumps(expected_post_data),
                headers=expected_headers,
            )

    @patch.object(AiTranslationService, "client", PropertyMock())
    @patch("translatable_xblocks.ai_translations_service.TieredCache")
    def test_translate_base64_image(self, mock_cache):
        # Given a valid translations service
        ai_service = AiTranslationService()

        # ... and a cache miss for requested data
        mock_cache.get_cached_response = Mock(
            return_value=CachedResponse(False, "foo", "****")
        )
        mock_cache.set_all_tiers = Mock()

        with patch.object(ai_service.client, "post") as mock_post:
            translated_html = (
                "<html><body>"
                "<img src='BASE64_IMG_PLACEHOLDER_0' alt='Translated Example'><body></html>"
            )

            mock_post.return_value.status_code = 200
            mock_post.return_value.json.return_value = {
                "translated_content": translated_html
            }

            content = (
                "<html><body>"
                "<img src='data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUA' alt='Example'><body></html>"
            )
            source_language = "en"
            target_language = "es"
            block_id = "block-v1:foo+bar"
            user_id = "1"

            ai_service.translate(
                content,
                source_language,
                target_language,
                block_id,
                user_id,
                mimetype="application/xml",
            )

            expected_url = "example.com/api/translate-xblock/"
            expected_post_data = {
                "block_id": block_id,
                "source_language": source_language,
                "target_language": target_language,
                "content": (
                    "<html><body>"
                    "<img src=\"BASE64_IMG_PLACEHOLDER_0\" alt='Example'><body></html>"
                ),
                "content_hash": sha256(content.encode("utf-8")).hexdigest(),
                "cache_override": "none",
            }
            expected_headers = {
                "content-type": "application/json",
            }

            # Then I call the endpoint with the correct data
            mock_post.assert_called_with(
                expected_url,
                data=json.dumps(expected_post_data),
                headers=expected_headers,
            )

    @patch.object(AiTranslationService, "client", PropertyMock())
    @patch("translatable_xblocks.ai_translations_service.TieredCache")
    def test_translate_xml_cached(self, mock_cache):
        # Given a valid translations service
        ai_service = AiTranslationService()
        original_content = (
            '<option correct="false">Attributes should be quoted</option>'
        )

        # ... and a cache hit for requested data
        transformed_and_cached_translation_content = original_content
        mock_cache.get_cached_response = Mock(
            return_value=CachedResponse(
                True, "foo", transformed_and_cached_translation_content
            )
        )
        mock_cache.set_all_tiers = Mock()

        with patch.object(ai_service.client, "post") as mock_post:
            content = original_content
            source_language = "en"
            target_language = "es"
            block_id = "block-v1:foo+bar"
            user_id = "1"

            # When I call "translate" for XML by specifying an XML mimetype
            translated_content = ai_service.translate(
                content,
                source_language,
                target_language,
                block_id,
                user_id,
                mimetype="application/xml",
            )

            # Then the Translations IDA is NOT called
            mock_post.assert_not_called()
            self.assertEqual(
                translated_content, transformed_and_cached_translation_content
            )

    @patch("translatable_xblocks.ai_translations_service.TieredCache")
    @patch.object(AiTranslationService, "client", PropertyMock())
    def test_get_course_translation_languages_cache_hit(self, mock_cache):
        # Given a valid translations service
        ai_service = AiTranslationService()
        valid_course_key = "course-v1:edX+DemoX+Demo_Course"
        cache_key = ai_service.config_cache_key(valid_course_key)

        # ... and a cache hit for requested data
        mock_cached_data = [{"code": "es", "label": "Spanish", "enabled": True}]

        mock_cache.get_cached_response = Mock(
            return_value=CachedResponse(True, cache_key, mock_cached_data)
        )
        mock_cache.set_all_tiers = Mock()

        # When I call "get_course_translation_languages" with a cache hit
        with patch.object(ai_service.client, "get") as mock_get:
            response = ai_service.get_course_translation_languages(valid_course_key)

            # Then I should get cached data back
            self.assertEqual(response, mock_cached_data)

            # ... without calling AI Translations
            mock_get.assert_not_called()

    @patch("translatable_xblocks.ai_translations_service.TieredCache")
    @patch.object(AiTranslationService, "client", PropertyMock())
    def test_get_course_translation_languages_cache_miss(self, mock_cache):
        # Given a valid translations service
        ai_service = AiTranslationService()
        valid_course_key = "course-v1:edX+DemoX+Demo_Course"
        cache_key = ai_service.config_cache_key(valid_course_key)

        # ... and a cache miss for requested data
        mock_cache.get_cached_response = Mock(
            return_value=CachedResponse(False, cache_key, None)
        )

        with patch.object(ai_service.client, "get") as mock_get:
            mock_get.return_value.status_code = 200
            mock_get.return_value.json.return_value = {
                "course_id": valid_course_key,
                "available_unit_translation_languages": [
                    {"code": "es", "label": "Spanish", "enabled": True}
                ],
            }

            expected_url = "example.com/api/available_languages/"
            expected_response_data = [
                {"code": "es", "label": "Spanish", "enabled": True}
            ]
            expected_headers = {
                "content-type": "application/json",
            }

            # When I get the course translation languages
            response = ai_service.get_course_translation_languages(valid_course_key)

            # Then I call the correct endpoint...
            mock_get.assert_called_with(
                expected_url,
                params={"course_id": valid_course_key},
                headers=expected_headers,
                timeout=15,
            )

            # ... getting the expected data
            self.assertEqual(response, expected_response_data)

            # ... and setting the cached value
            mock_cache.set_all_tiers.assert_called_once_with(
                cache_key,
                expected_response_data,
                settings.CONFIG_REQUEST_CACHE_TTL,
            )

    @patch("translatable_xblocks.ai_translations_service.TieredCache")
    @patch.object(AiTranslationService, "client", PropertyMock())
    def test_get_course_translation_languages_failure(self, mock_cache):
        # Given a valid translations service
        ai_service = AiTranslationService()
        valid_course_key = "course-v1:edX+DemoX+Demo_Course"
        cache_key = ai_service.config_cache_key(valid_course_key)

        # ... and a cache miss for requested data
        mock_cache.get_cached_response = Mock(
            return_value=CachedResponse(False, cache_key, None)
        )

        with patch.object(ai_service.client, "get") as mock_get:
            mock_get.return_value.status_code = 400

            expected_url = "example.com/api/available_languages/"
            expected_headers = {
                "content-type": "application/json",
            }

            # When get_course_translation_languages fails, it raises an exception
            with self.assertRaises(Exception):
                ai_service.get_course_translation_languages(valid_course_key)

                mock_get.assert_called_with(
                    expected_url,
                    params={"course_id": valid_course_key},
                    headers=expected_headers,
                )

            # And sets the error cache
            mock_cache.set_all_tiers.assert_called_once_with(
                cache_key, None, settings.REQUEST_CACHE_FAILURE_TIMEOUT
            )

    @patch("translatable_xblocks.ai_translations_service.TieredCache")
    @patch.object(AiTranslationService, "client", PropertyMock())
    def test_get_course_translation_languages_cached_failure(self, mock_cache):
        # Given a valid translations service
        ai_service = AiTranslationService()
        valid_course_key = "course-v1:edX+DemoX+Demo_Course"
        cache_key = ai_service.config_cache_key(valid_course_key)

        # ... and a cached failure for requested data
        mock_cache.get_cached_response = Mock(
            return_value=CachedResponse(True, cache_key, None)
        )

        with patch.object(ai_service.client, "get") as mock_get:
            # When I get_course_translation_languages...
            mock_get.return_value.status_code = 400

            # Then I get a chached response
            with self.assertRaises(Exception):
                ai_service.get_course_translation_languages(valid_course_key)

                mock_cache.assert_called_once_with(cache_key)

            # And I do not call AI Translations
            mock_get.assert_not_called()

    @patch("translatable_xblocks.ai_translations_service.TieredCache")
    @patch.object(AiTranslationService, "client", PropertyMock())
    def test_post_course_translation_languages(self, mock_cache):
        # Given a valid translations service
        ai_service = AiTranslationService()
        valid_course_key = "course-v1:edX+DemoX+Demo_Course"
        cache_key = ai_service.config_cache_key(valid_course_key)

        with patch.object(ai_service.client, "post") as mock_post:
            mock_post.return_value.status_code = 200
            mock_post.return_value.json.return_value = {
                "course_id": valid_course_key,
                "available_unit_translation_languages": [
                    {"code": "es", "label": "Spanish", "enabled": True}
                ],
            }

            expected_url = "example.com/api/available_languages/"
            expected_post_data = {
                "course_id": valid_course_key,
                "available_unit_translation_languages": [
                    {"code": "es", "label": "Spanish", "enabled": True}
                ],
            }
            expected_headers = {
                "content-type": "application/json",
            }

            # When I change the course translation languages
            ai_service.post_course_translation_languages(expected_post_data)

            # Then I call the expected endpoint
            mock_post.assert_called_with(
                expected_url,
                data=json.dumps(expected_post_data),
                headers=expected_headers,
            )

            # ... and update my cached values
            mock_cache.set_all_tiers.assert_called_once_with(
                cache_key,
                expected_post_data["available_unit_translation_languages"],
                settings.CONFIG_REQUEST_CACHE_TTL,
            )

    @patch("translatable_xblocks.ai_translations_service.TieredCache")
    @patch.object(AiTranslationService, "client", PropertyMock())
    def test_post_course_translation_languages_fail(self, mock_cache):
        # Given a valid translations service
        ai_service = AiTranslationService()
        valid_course_key = "course-v1:edX+DemoX+Demo_Course"

        with patch.object(ai_service.client, "post") as mock_post:
            mock_post.return_value.status_code = 400
            mock_post.return_value.json.return_value = {
                "course_id": valid_course_key,
                "available_unit_translation_languages": [
                    {"code": "es", "label": "Spanish", "enabled": True}
                ],
            }

            expected_url = "example.com/api/available_languages/"
            expected_post_data = {
                "course_id": valid_course_key,
                "available_unit_translation_languages": [
                    {"code": "es", "label": "Spanish", "enabled": True}
                ],
            }
            expected_headers = {
                "content-type": "application/json",
            }

            # When I try to change course translation languages and it fails, it raises an exception
            with self.assertRaises(Exception):
                ai_service.post_course_translation_languages(expected_post_data)

                mock_post.assert_called_with(
                    expected_url,
                    data=json.dumps(expected_post_data),
                    headers=expected_headers,
                )

            # ... and does not update cached values
            mock_cache.assert_not_called()



"""Config specific to the developer environment"""


def plugin_settings(settings):
    """
    App-specific settings
    """
    # AI Translations Service
    settings.AI_TRANSLATIONS_API_URL = "http://host.docker.internal:18760"
    settings.AI_TRANSLATIONS_OAUTH_APP_NAME = "ai_translations-backend-service"

    # Add filter config
    settings.OPEN_EDX_FILTERS_CONFIG = getattr(settings, "OPEN_EDX_FILTERS_CONFIG", {})

    settings.OPEN_EDX_FILTERS_CONFIG[
        "org.openedx.learning.xblock.render.started.v1"
    ] = {
        "fail_silently": False,
        "pipeline": ["translatable_xblocks.filters.UpdateRequestLanguageCode"],
    }

    # Cache Settings (shorter, ease of development)

    # TTL in seconds for successful request to store in cache (5 minutes)
    settings.REQUEST_CACHE_SUCCESS_TIMEOUT = 300

    # TTL in seconds for successful config request to store in cache (1 minute)
    settings.CONFIG_REQUEST_CACHE_TTL = 60

    # TIL in seconds for failure request to store in cache (30 seconds)
    settings.REQUEST_CACHE_FAILURE_TIMEOUT = 30

    # Timeout in seconds for API requests
    settings.AI_TRANSLATIONS_API_TIMEOUT = 15



"""Tests for TranslatableXBlock base behavior"""

from unittest import TestCase
from unittest.mock import Mock, patch

import ddt

from translatable_xblocks.base import TranslatableXBlock


class MockSuperXBlock:
    """Mock XBlock for layering on translated behavior"""

    # pylint: disable=unused-argument
    def student_view(self, context, **kwargs):
        return self

    # pylint: disable=unused-argument
    def handle_ajax(self, context, **kwargs):
        return self


class MockTranslatedXBlock(TranslatableXBlock, MockSuperXBlock):
    """Mock Translated XBlock to add overridden tranlsation behavior."""


@ddt.ddt
class TestTranslatableXBlockBase(TestCase):
    """Tests for TranslatableXBlock base behavior"""

    def setUp(self):
        """Create a test block."""

        # Initialize the block
        self.scope_ids = Mock()
        self.runtime = Mock()
        self.xblock = MockTranslatedXBlock(self.runtime, scope_ids=self.scope_ids)

        # Initialize some block fields
        self.xblock.ai_translations_service = Mock()
        self.xblock.location = Mock()

        return super().setUp()

    @patch(
        "translatable_xblocks.base.TranslatableXBlocksFeatureConfig.xpert_translations_enabled_for_course"
    )
    def test_should_translate_not_enabled(self, translations_enabled):
        # Given translations are not enabled for a given scope
        translations_enabled.return_value = False

        # When I ask if we should translate
        should_translate = self.xblock.should_translate

        # Then we return False
        self.assertFalse(should_translate)

    @ddt.unpack
    @ddt.data(["en", None], [None, "ar"])
    @patch(
        "translatable_xblocks.base.TranslatableXBlocksFeatureConfig.xpert_translations_enabled_for_course"
    )
    def test_should_translate_no_language(
        self, source_lang, translate_lang, translations_enabled
    ):
        # Given translations are enabled
        translations_enabled.return_value = True

        # ... but a block isn't given a source / translate langauge
        self.xblock.source_lang = source_lang
        self.xblock.translate_lang = translate_lang

        # When I ask if we should translate
        should_translate = self.xblock.should_translate

        # Then we return False
        self.assertFalse(should_translate)

    @patch(
        "translatable_xblocks.base.TranslatableXBlocksFeatureConfig.xpert_translations_enabled_for_course"
    )
    def test_should_translate_same_language(self, translations_enabled):
        # Given translations are enabled
        translations_enabled.return_value = True

        # ... but the selected languages are the same
        self.xblock.source_lang = "en"
        self.xblock.translate_lang = "en"
        # When I ask if we should translate
        should_translate = self.xblock.should_translate

        # Then we return False
        self.assertFalse(should_translate)

    @patch(
        "translatable_xblocks.base.TranslatableXBlocksFeatureConfig.xpert_translations_enabled_for_course"
    )
    def test_should_translate(self, translations_enabled):
        # Given translations are enabled
        translations_enabled.return_value = True

        # ... and languages are allowed and different
        self.xblock.source_lang = "en"
        self.xblock.translate_lang = "gf"

        # When I ask if we should translate
        should_translate = self.xblock.should_translate

        # Then we return True
        self.assertTrue(should_translate)

    @patch.object(TranslatableXBlock, "should_translate")
    def test_student_view_untranslated(
        self, mock_should_translate
    ):  # pylint: disable=unused-argument
        # Given a translatable XBlock
        # ... where a translation is not requested
        self.xblock.should_translate = False

        # When I call the student_view
        with patch.object(MockSuperXBlock, "student_view") as super_student_view:
            context = {}
            kwargs = {"foo": "bar", "baz": "buzz"}
            self.xblock.student_view(context, **kwargs)

            # Then I return the parent block's student_view
            super_student_view.assert_called_with(context, **kwargs)

            # ... and my source / target langs are updated
            self.assertIsNone(self.xblock.source_lang)
            self.assertIsNone(self.xblock.translate_lang)

    @patch("translatable_xblocks.base.user_is_admin")
    @patch.object(TranslatableXBlock, "should_translate")
    def test_student_view_translated(
        self, mock_should_translate, mock_user_is_admin
    ):  # pylint: disable=unused-argument
        # Given a translatable XBlock
        # ... where a translation is requested
        self.xblock.should_translate = True

        # When I call the student_view
        with patch.object(MockSuperXBlock, "student_view") as super_student_view:
            context = {"src_lang": "en", "dest_lang": "gf"}
            kwargs = {"foo": "bar", "baz": "buzz"}
            self.xblock.student_view(context, **kwargs)

            # Then I return the parent block's student_view
            super_student_view.assert_called_with(context, **kwargs)

            # ... and my source / target langs are updated
            self.assertEqual(self.xblock.source_lang, "en")
            self.assertEqual(self.xblock.translate_lang, "gf")

    @patch("translatable_xblocks.base.user_is_admin")
    @patch.object(TranslatableXBlock, "should_translate")
    def test_student_view_translated_cache_reset(
        self, mock_should_translate, mock_user_is_admin
    ):  # pylint: disable=unused-argument
        # Given a translatable XBlock
        # ... where a translation is requested
        self.xblock.should_translate = True

        # When I call the student_view
        with patch.object(MockSuperXBlock, "student_view") as super_student_view:
            # ... and a cache reset is requested
            mock_user_is_admin.return_value = True
            cache_reset = {"xt_cache": "reset"}

            context = {"src_lang": "en", "dest_lang": "gf", **cache_reset}
            kwargs = {"foo": "bar", "baz": "buzz"}
            self.xblock.student_view(context, **kwargs)

            # Then I return the parent block's student_view
            super_student_view.assert_called_with(context, **kwargs)

            # ... and my source / target langs are updated
            self.assertEqual(self.xblock.source_lang, "en")
            self.assertEqual(self.xblock.translate_lang, "gf")

            # ... and cache behavior is reset after student_view
            self.assertIsNone(self.xblock.xt_cache_behavior)

    @patch.object(TranslatableXBlock, "should_translate")
    def test_handle_ajax_untranslated(
        self, mock_should_translate
    ):  # pylint: disable=unused-argument
        # Given a translatable XBlock
        # ... where a translation is NOT requested
        self.xblock.should_translate = False

        # When I call handle_ajax
        with patch.object(MockSuperXBlock, "handle_ajax") as super_handle_ajax:
            with patch.object(self.xblock, "runtime") as mock_runtime:
                dispatch = "request-name"
                data = {"foo": "bar"}
                self.xblock.handle_ajax(dispatch, data)

                # Then I call the parent block's handle_ajax with expected data
                super_handle_ajax.assert_called_once_with(dispatch, data)

                # ... and do not activate the i18n language for frontend translations
                mock_runtime.service(
                    self.xblock, "i18n"
                ).translator.activate.assert_not_called()

    @patch.object(TranslatableXBlock, "should_translate")
    def test_handle_ajax_translated(
        self, mock_should_translate
    ):  # pylint: disable=unused-argument
        # Given a translatable XBlock
        # ... where a translation is requested
        self.xblock.should_translate = True
        self.xblock.translate_lang = "gf"

        # When I call handle_ajax
        with patch.object(MockSuperXBlock, "handle_ajax") as super_handle_ajax:
            with patch.object(self.xblock, "runtime") as mock_runtime:
                dispatch = "request-name"
                data = {"foo": "bar"}
                self.xblock.handle_ajax(dispatch, data)

                # Then I call the parent block's handle_ajax with expected data
                super_handle_ajax.assert_called_once_with(dispatch, data)

                # ... and activate the i18n language for frontend translations
                mock_runtime.service(
                    self.xblock, "i18n"
                ).translator.activate.assert_called_with(self.xblock.translate_lang)



"""
Translatable version of edx-platform/xmodule/video_block
"""

from xblock.core import XBlock
from xblock.fields import Scope
from xmodule.video_block import VideoBlock as BaseVideoBlock

from translatable_xblocks.base import TranslatableXBlock
from translatable_xblocks.fields import TranslatableString

# Make '_' a no-op so we can scrape strings. Using lambda instead of
#  `django.utils.translation.gettext_noop` because Django cannot be imported in this file
_ = lambda text: text  # pylint:  disable=unnecessary-lambda-assignment


@XBlock.wants("settings", "completion", "i18n", "request_cache")
@XBlock.needs("mako", "user")
class VideoBlock(TranslatableXBlock, BaseVideoBlock):
    """
    Only translate display_name which is plain text
    """

    display_name = TranslatableString(
        help=_("The display name for this component."),
        display_name=_("Component Display Name"),
        default="Video",
        scope=Scope.settings,
    )



"""Tests for translatable XBlock fields."""

import ast
from unittest import TestCase
from unittest.mock import Mock, PropertyMock, patch

from xblock.core import XBlock
from xblock.field_data import DictFieldData

from translatable_xblocks.base import TranslatableXBlock
from translatable_xblocks.fields import (
    TranslatableHTML,
    TranslatableList,
    TranslatableString,
    TranslatableXMLString,
)


class MockTranslationService:
    """Mock translation service for testing."""

    # pylint: disable=unused-argument, too-many-positional-arguments
    def translate(
        self,
        text,
        source_lang,
        target_lang,
        location,
        user_id,
        cache_override=None,
        mimetype=None,
    ):
        """Mock implementation of the translate function."""

        # For JSON data, mock translating of json strings, but not keys
        if mimetype == "application/json":
            list_data = ast.literal_eval(text)
            translated_list = []

            for item in list_data:
                # String items directly are translated
                if isinstance(item, str):
                    translated_list.append(item)

                # Where lists are tuples (e.g. the Poll XBlock), they represent an ordered dict.
                if isinstance(item, tuple):
                    translated_tuple = []

                    for sub_item in item:
                        # A string item is a key and should not be translated
                        if isinstance(sub_item, str):
                            translated_tuple.append(sub_item)
                        # A dict should have its values, but not keys, translated
                        elif isinstance(sub_item, dict):
                            translated_tuple.append(
                                {k: f"Translated({v})" for k, v in sub_item.items()}
                            )
                    translated_list.append(tuple(translated_tuple))
            # Return value from AI Translations is a string
            return str(translated_list)

        # For HTML data, wrap in "TranslatedHTML()"
        if mimetype == "text/html":
            return f"TranslatedHTML({text})"

        # For XML data, wrap in "TranslatedXML()"
        if mimetype == "text/xml":
            return f"TranslatedXML({text})"

        # Default mock translation: wrap in "Translated()"
        return f"Translated({text})"


class TestTranslatableHTML(TestCase):
    """Tests for TranslatableHTML field."""

    @XBlock.needs("settings")
    class MockXBlock(TranslatableXBlock):
        """A mock XBlock for testing TranslatableHTML fields."""

        html_field = TranslatableHTML(default="<p>HTML Field</p>")

    def setUp(self):
        """Create a test XBlock with minimal dependencies."""
        self.scope_ids = Mock()
        self.runtime = Mock()

        # Use DictFieldData to simulate real field data behavior
        self.original_html_content = "<p>Foo Bar</p>"
        field_data = DictFieldData(
            {
                "html_field": self.original_html_content,
            }
        )

        self.block = self.MockXBlock(
            self.runtime, field_data=field_data, scope_ids=self.scope_ids
        )

        # Required fields for translation
        self.block.location = Mock()
        self.block.ai_translations_service = MockTranslationService()

        super().setUp()

    @patch.object(MockXBlock, "should_translate", new_callable=PropertyMock)
    def test_translate_html(self, mock_should_translate):
        # Given an XBlock with TranslatableHTML field where we *should* translate.
        mock_should_translate.return_value = True

        # When I access the field
        translated = self.block.html_field

        # Then the contents is translated
        self.assertEqual(translated, f"TranslatedHTML({self.original_html_content})")

    @patch.object(MockXBlock, "should_translate", new_callable=PropertyMock)
    def test_no_translate_html(self, mock_should_translate):
        # Given an XBlock with TranslatableHTML field where we *should not* translate.
        mock_should_translate.return_value = False

        # When I access the field
        translated = self.block.html_field

        # Then the contents is not translated
        self.assertEqual(translated, self.original_html_content)


class TestTranslatableString(TestCase):
    """Tests for TranslatableString field."""

    @XBlock.needs("settings")
    class MockXBlock(TranslatableXBlock):
        """A mock XBlock for testing TranslatableString fields."""

        string_field = TranslatableString(default="String Field")

    def setUp(self):
        """Create a test XBlock with minimal dependencies."""
        self.scope_ids = Mock()
        self.runtime = Mock()

        # Use DictFieldData to simulate real field data behavior
        self.original_string_content = "Foo Bar"
        field_data = DictFieldData(
            {
                "string_field": self.original_string_content,
            }
        )

        self.block = self.MockXBlock(
            self.runtime, field_data=field_data, scope_ids=self.scope_ids
        )

        # Required fields for translation
        self.block.location = Mock()
        self.block.ai_translations_service = MockTranslationService()

        super().setUp()

    @patch.object(MockXBlock, "should_translate", new_callable=PropertyMock)
    def test_translate_string(self, mock_should_translate):
        # Given an XBlock with TranslatableString field where we *should* translate.
        mock_should_translate.return_value = True

        # When I access the field
        translated = self.block.string_field

        # Then the contents is translated
        self.assertEqual(translated, f"Translated({self.original_string_content})")

    @patch.object(MockXBlock, "should_translate", new_callable=PropertyMock)
    def test_no_translate_string(self, mock_should_translate):
        # Given an XBlock with TranslatableString field where we *should not* translate.
        mock_should_translate.return_value = False

        # When I access the field
        translated = self.block.string_field

        # Then the contents is not translated
        self.assertEqual(translated, self.original_string_content)


class TestTranslatableXMLString(TestCase):
    """Tests for TranslatableXMLString field."""

    @XBlock.needs("settings")
    class MockXBlock(TranslatableXBlock):
        """A mock XBlock for testing TranslatableXMLString fields."""

        xml_field = TranslatableXMLString(default="<field>Value</field>")

    def setUp(self):
        """Create a test XBlock with minimal dependencies."""
        self.scope_ids = Mock()
        self.runtime = Mock()

        # Use DictFieldData to simulate real field data behavior
        self.original_xml_content = "<field>Foo Bar</field>"
        field_data = DictFieldData(
            {
                "xml_field": self.original_xml_content,
            }
        )

        self.block = self.MockXBlock(
            self.runtime, field_data=field_data, scope_ids=self.scope_ids
        )

        # Required fields for translation
        self.block.location = Mock()
        self.block.ai_translations_service = MockTranslationService()

        super().setUp()

    @patch.object(MockXBlock, "should_translate", new_callable=PropertyMock)
    def test_translate_xml_string(self, mock_should_translate):
        # Given an XBlock with TranslatableXMLString field where we *should* translate.
        mock_should_translate.return_value = True

        # When I access the field
        translated = self.block.xml_field

        # Then the contents is translated
        self.assertEqual(translated, f"TranslatedXML({self.original_xml_content})")

    @patch.object(MockXBlock, "should_translate", new_callable=PropertyMock)
    def test_no_translate_xml_string(self, mock_should_translate):
        # Given an XBlock with TranslatableXMLString field where we *should not* translate.
        mock_should_translate.return_value = False

        # When I access the field
        translated = self.block.xml_field

        # Then the contents is not translated
        self.assertEqual(translated, self.original_xml_content)


class TestTranslatableList(TestCase):
    """Tests for TranslatableList field."""

    @XBlock.needs("settings")
    class MockXBlock(TranslatableXBlock):
        """A mock XBlock for testing TranslatableXList fields."""

        # Example list field for how it's used in poll XBlocks
        list_field = TranslatableList(default=["key", {"label": "value"}])

    def setUp(self):
        """Create a test XBlock with minimal dependencies."""
        self.scope_ids = Mock()
        self.runtime = Mock()

        # Use DictFieldData to simulate real field data behavior
        self.original_list_content = [
            ("A", {"label": "Apple"}),
            ("B", {"label": "Banana"}),
            ("C", {"label": "Cherry"}),
        ]
        field_data = DictFieldData(
            {
                "list_field": self.original_list_content,
            }
        )

        self.block = self.MockXBlock(
            self.runtime, field_data=field_data, scope_ids=self.scope_ids
        )

        # Required fields for translation
        self.block.location = Mock()
        self.block.ai_translations_service = MockTranslationService()

        super().setUp()

    @patch.object(MockXBlock, "should_translate", new_callable=PropertyMock)
    def test_translate_list(self, mock_should_translate):
        # Given an XBlock with TranslatableList field where we *should* translate.
        mock_should_translate.return_value = True

        # When I access the field
        translated = self.block.list_field

        # Then the contents is translated
        self.assertEqual(
            translated,
            [
                ("A", {"label": "Translated(Apple)"}),
                ("B", {"label": "Translated(Banana)"}),
                ("C", {"label": "Translated(Cherry)"}),
            ],
        )

    @patch.object(MockXBlock, "should_translate", new_callable=PropertyMock)
    def test_no_translate_list(self, mock_should_translate):
        # Given an XBlock with TranslatableList field where we *should not* translate.
        mock_should_translate.return_value = False

        # When I access the field
        translated = self.block.list_field

        # Then the contents is not translated
        self.assertEqual(translated, self.original_list_content)



"""
Translatable XBlock base class & utils.
"""

# pylint:  disable=no-member, unnecessary-lambda-assignment

import logging

from django.utils.translation import to_locale
from xblock.core import XBlock
from xblock.fields import Scope, String

from translatable_xblocks.ai_translations_service import AiTranslationService
from translatable_xblocks.config import TranslatableXBlocksFeatureConfig
from translatable_xblocks.platform_imports import user_is_admin

logger = logging.getLogger(__name__)

# Make '_' a no-op so we can scrape strings. Using lambda instead of
#  `django.utils.translation.gettext_noop` because Django cannot be imported in this file
_ = lambda text: text


class TranslatableXBlock(XBlock):
    """
    Base block for translatable XBlocks.

    See docs/how-tos/add_a_new_xblock.rst for usage instructions.
    """

    # Translations service
    ai_translations_service = AiTranslationService()

    source_lang = String(
        display_name=_("Source Language"),
        help=_("The source language of this content."),
        scope=Scope.user_state,
    )

    translate_lang = String(
        display_name=_("Translation Language"),
        help=_("The language to translate this block into."),
        scope=Scope.user_state,
    )

    xt_cache_behavior = None

    @property
    def should_translate(self):
        """
        Determine if we should translate for this context.

        We should only translate if:
        1) Xpert Tranlsations feature is enabled for this scope.
        2) Page has a source and translate language.
        3) The language codes differ.

        NOTE - We don't check here if the languages are allowed by our course config.
        These are checked by AI Translations, which will send an error if a bad language
        code is provided.
        """
        if not TranslatableXBlocksFeatureConfig.xpert_translations_enabled_for_course(
            self.location.course_key
        ):
            return False

        if not (self.source_lang and self.translate_lang):
            return False

        return self.source_lang != self.translate_lang

    def student_view(self, context, **kwargs):
        """
        Pull translation-specific values from request context, then call parent rendering.
        """
        # Get translation source / destination language
        self.source_lang = context.get("src_lang")
        self.translate_lang = context.get("dest_lang")

        # If translation behavior is enabled / selected...
        if self.should_translate:
            # Get cache behavior overrides, only if user is admin
            user = self.runtime.service(self, "user").get_current_user()
            self.xt_cache_behavior = (
                context.get("xt_cache") if user_is_admin(user) else None
            )

            # Activate the language for this request, enables template translations
            self.runtime.service(self, "i18n").translator.activate(
                to_locale(self.translate_lang)
            )

            # Call original student_view function
            fragment = super().student_view(context, **kwargs)

            # Reset cache override requests
            self.xt_cache_behavior = None

            return fragment

        # ... otherwise, just call original student_view
        return super().student_view(context, **kwargs)

    def handle_ajax(self, dispatch, data):
        """
        Handle AJAX calls for XBlocks, called by courseware.block_render.

        Since we patch template translations per request, this other async way of getting
        a view needs to also be patched.
        """
        # If translation behavior is enabled / selected...
        if self.should_translate:
            # Activate the language for this request, enables template translations
            self.runtime.service(self, "i18n").translator.activate(
                to_locale(self.translate_lang)
            )

        return super().handle_ajax(dispatch, data)



"""
Utilities used across the Translatable XBlocks Repo.
"""

import re


def replace_img_base64_with_placeholder(content):
    """
    Replace base64 images with placeholder.

    This is needed to prevent the code from sending huge amount of text to Google
    as base64 images are usually single long strings that exceeds Google translate limit.
    """
    base64_images = []
    placeholder_template = "BASE64_IMG_PLACEHOLDER_{index}"

    # Function to replace base64 src with placeholders
    def replace_match(match):
        base64_images.append((match.group(3), match.group(4)))
        placeholder = placeholder_template.format(index=len(base64_images) - 1)
        return f'<img {match.group(1)}src="{placeholder}"'

    # Regex to find <img> tags with base64 src
    img_tag_regex = (
        r'<img\s+([^>]*?)src=(["\'])data:image/(png|jpeg|jpg|gif);base64,([^"\']+)\2'
    )
    processed_content = re.sub(img_tag_regex, replace_match, content)

    return processed_content, base64_images


def reinsert_base64_images(content, base64_images):
    """Replace placeholders with the original base64 string."""
    for index, (image_type, base64_str) in enumerate(base64_images):
        placeholder = f"BASE64_IMG_PLACEHOLDER_{index}"
        content = content.replace(
            placeholder, f"data:image/{image_type};base64,{base64_str}"
        )
    return content



"""
Translatable versions of XBlock fields.
"""

import ast

from xblock.fields import List, String, XMLString

from translatable_xblocks.base import TranslatableXBlock


class TranslatableHTML(String):
    """
    Translatable HTML field rich content (HTML/mixed markup).

    explicitly sends mimetype='text/html' to translation service
    """

    def xt_translate(self, value, source_lang, translate_lang, xblock):
        """Translate a value to a given language. Function, defined here, allows for overriding per block."""
        return xblock.ai_translations_service.translate(
            value,
            source_lang,
            translate_lang,
            xblock.location,
            xblock.scope_ids.user_id,
            cache_override=xblock.xt_cache_behavior,
            mimetype="text/html",
        )

    def __get__(self, xblock, xblock_class):
        """
        Get the value of a string and return the translated value, if applicable.
        """
        original_value = super().__get__(xblock, xblock_class)

        # This allows get on initial definition without breaks
        if xblock is None:
            return self

        # Only operate on classes that subclass from TranslatableXBlock
        if not issubclass(xblock_class, TranslatableXBlock):
            raise AssertionError(
                f"{xblock_class} is not subclass of TranslatableXblock"
            )

        if xblock.should_translate:
            translated_value = self.xt_translate(
                original_value, xblock.source_lang, xblock.translate_lang, xblock
            )
            return translated_value

        return original_value


class TranslatableXMLString(XMLString):
    """
    Translatable String field on an XBlock with extra handling for XML processing.

    Used for fields (e.g. CAPA / Problem) that have XML definitions.
    """

    def xt_translate(self, value, source_lang, translate_lang, xblock):
        """Override translation function for XML."""
        return xblock.ai_translations_service.translate(
            value,
            source_lang,
            translate_lang,
            xblock.location,
            xblock.scope_ids.user_id,
            mimetype="text/xml",
            cache_override=xblock.xt_cache_behavior,
        )

    def __get__(self, xblock, xblock_class):
        """
        Get the value of a string and return the translated value, if applicable.
        """
        original_value = super().__get__(xblock, xblock_class)

        # This allows get on initial definition without breaks
        if xblock is None:
            return self

        # Only operate on classes that subclass from TranslatableXBlock
        if not issubclass(xblock_class, TranslatableXBlock):
            raise AssertionError(
                f"{xblock_class} is not subclass of TranslatableXblock"
            )

        if xblock.should_translate:
            translated_value = self.xt_translate(
                original_value, xblock.source_lang, xblock.translate_lang, xblock
            )
            return translated_value

        return original_value


class TranslatableString(String):
    """
    Translatable String field on an plaine text content.

    explicitly sends mimetype='text/plain' to translation service
    """

    def xt_translate(self, value, source_lang, translate_lang, xblock):
        """Translate a value to a given language. Function, defined here, allows for overriding per block."""
        return xblock.ai_translations_service.translate(
            value,
            source_lang,
            translate_lang,
            xblock.location,
            xblock.scope_ids.user_id,
            cache_override=xblock.xt_cache_behavior,
            mimetype="text/plain",
        )

    def __get__(self, xblock, xblock_class):
        """
        Get the value of a string and return the translated value, if applicable.
        """
        original_value = super().__get__(xblock, xblock_class)

        # This allows get on initial definition without breaks
        if xblock is None:
            return self

        # Only operate on classes that subclass from TranslatableXBlock
        if not issubclass(xblock_class, TranslatableXBlock):
            raise AssertionError(
                f"{xblock_class} is not subclass of TranslatableXblock"
            )

        if xblock.should_translate:
            translated_value = self.xt_translate(
                original_value, xblock.source_lang, xblock.translate_lang, xblock
            )
            return translated_value

        return original_value


class TranslatableList(List):
    """A translatable variant of a List with JSON items."""

    def xt_translate(self, value, source_lang, translate_lang, xblock):
        """Translate a value to a given language. Function, defined here, allows for overriding per block."""
        # Encode the list as a string.
        list_as_string = str(value)

        # Translate the stringified list / JSON using AI Translations
        translated_list_string = xblock.ai_translations_service.translate(
            list_as_string,
            source_lang,
            translate_lang,
            xblock.location,
            xblock.scope_ids.user_id,
            cache_override=xblock.xt_cache_behavior,
            mimetype="application/json",
        )

        # Convert the translated JSON object back to a Python list
        translated_list = ast.literal_eval(translated_list_string)
        return translated_list

    def __get__(self, xblock, xblock_class):
        """
        Get the value of a string and return the translated value, if applicable.
        """
        original_value = super().__get__(xblock, xblock_class)

        # This allows get on initial definition without breaks
        if xblock is None:
            return self

        # Only operate on classes that subclass from TranslatableXBlock
        if not issubclass(xblock_class, TranslatableXBlock):
            raise AssertionError(
                f"{xblock_class} is not subclass of TranslatableXblock"
            )

        if xblock.should_translate:
            translated_value = self.xt_translate(
                original_value, xblock.source_lang, xblock.translate_lang, xblock
            )
            return translated_value

        return original_value


TRANSLATABLE_FIELDS = [
    TranslatableHTML,
    TranslatableXMLString,
    TranslatableString,
    TranslatableList,
]



"""
Translatable version of edx-platform/xmodule.capa_block
"""

# pylint:  disable=unnecessary-lambda-assignment

import logging

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from xblock.core import XBlock
from xblock.fields import Scope
from xmodule.capa_block import ProblemBlock as BaseProblemBlock

from translatable_xblocks.base import TranslatableXBlock
from translatable_xblocks.config import TranslatableXBlocksFeatureConfig
from translatable_xblocks.fields import TranslatableString, TranslatableXMLString

logger = logging.getLogger(__name__)

# Make '_' a no-op so we can scrape strings. Using lambda instead of
#  `django.utils.translation.gettext_noop` because Django cannot be imported in this file
_ = lambda text: text


try:
    FEATURES = getattr(settings, "FEATURES", {})
except ImproperlyConfigured:
    FEATURES = {}


@XBlock.needs("user")
@XBlock.needs("i18n")
@XBlock.needs("mako")
@XBlock.needs("cache")
@XBlock.needs("sandbox")
@XBlock.needs("replace_urls")
@XBlock.wants("call_to_action")
class ProblemBlock(TranslatableXBlock, BaseProblemBlock):
    """
    Our version of the ProblemBlock with added translation logic.
    data is text/xml(application/xml) and display_name is text/plain content(mimetype)
    """

    display_name = TranslatableString(
        display_name=_("Display Name"),
        help=_("The display name for this component."),
        scope=Scope.settings,
        # it'd be nice to have a useful default but it screws up other things; so,
        # use display_name_with_default for those
        default=_("Blank Problem"),
    )

    data = TranslatableXMLString(
        help=_("XML data for the problem"),
        scope=Scope.content,
        enforce_type=FEATURES.get("ENABLE_XBLOCK_XML_VALIDATION", True),
        default="<problem></problem>",
    )

    @property
    def should_translate_feedback(self):
        """
        Determine if we should translate feedback for this context.
        """
        return (
            TranslatableXBlocksFeatureConfig.problem_block_feedback_translation_enabled(
                self.location.course_key
            )
        )

    @property
    def has_lcp_error(self):
        """
        Checks if ProblemBlock will fail on accessing its LoncapaProblem
        """
        try:
            self.lcp
        except Exception:  # lint-amnesty, pylint: disable=broad-except
            logger.info(
                f"Feedback translation not available for ProblemBlock: {self.location}"
            )
            return True
        return False

    @property
    def has_file_submissions(self):
        """
        Checks if ProblemBlock's LoncapaProblem has file submissions in its responder
        """
        for responder in self.lcp.responders.values():
            if "filesubmission" in responder.allowed_inputfields:
                logger.info(
                    f"Feedback translation not available for ProblemBlock with FileSubmission: {self.location}"
                )
                return True
        return False

    def translate_feedback(self):
        """
        Translate feedback if we should, otherwise recalculate feedback in original language
        """
        # CorrectMap is a map between answer_id and response evaluation result
        student_answers_ids = list(self.lcp.correct_map.keys())

        # If student has attempted the problem block
        if self.attempts > 0 and student_answers_ids:
            # Get key from student answer
            student_answer_id = student_answers_ids[0]

            # Get feedback for student answer to translate
            feedback = self.lcp.correct_map.get_msg(student_answer_id)

            # If feedback is not empty
            if feedback:
                # If translation behavior is enabled / selected
                if self.should_translate:
                    # Get feedback translated value
                    translated_value = self.ai_translations_service.translate(
                        feedback,
                        self.source_lang,
                        self.translate_lang,
                        self.location,
                        self.scope_ids.user_id,
                    )

                    # Update Correct Map feedback with translated result
                    self.lcp.correct_map.set_property(
                        student_answer_id, "msg", translated_value
                    )
                else:
                    # Recalculate feedback as usual
                    self.update_correctness()

    def student_view(self, context, **kwargs):
        """
        Force feedback translation if should translate, then call parent rendering.
        """

        if self.should_translate_feedback:
            # Get translation source / destination language
            self.source_lang = context.get("src_lang")
            self.translate_lang = context.get("dest_lang")

            # Only translate feedback for supported use-cases
            feedback_translation_supported = not (
                self.has_lcp_error or self.has_file_submissions
            )

            if feedback_translation_supported:
                self.translate_feedback()

        return super().student_view(context, **kwargs)



"""
Waffle config for the AI Translations service.
"""

import logging

from translatable_xblocks.platform_imports import (
    course_waffle_flag,
    import_course_waffle_flag_override,
)

logger = logging.getLogger(__name__)

# Waffle settings
WAFFLE_NAMESPACE = "ai_translations"
LOG_PREFIX = "AI translations: "

# .. toggle_name: WHOLE_COURSE_TRANSLATIONS
# .. toggle_implementation: CourseWaffleFlag
# .. toggle_default: False
# .. toggle_description: Enabling this flag makes the course content translation
#   feature available. Works in conjunction with WHOLE_COURSE_TRANSLATIONS_ENABLED.
# .. toggle_warning: Requires ai-translations IDA to be available
# .. toggle_use_cases: opt_in
# .. toggle_creation_date: 2024-02-27
WHOLE_COURSE_TRANSLATIONS_FLAG = f"{WAFFLE_NAMESPACE}.whole_course_translations"

# .. toggle_name: WHOLE_COURSE_TRANSLATIONS_ENABLED
# .. toggle_implementation: CourseWaffleFlag
# .. toggle_default: False
# .. toggle_description: Enabling this feature allows course content to be sent to the
#   AI Translations service for automatic translation of course content.
# .. toggle_warning: Requires ai-translations IDA to be available
# .. toggle_use_cases: opt_in
# .. toggle_creation_date: 2024-05-09
WHOLE_COURSE_TRANSLATIONS_ENABLED_FLAG = (
    f"{WAFFLE_NAMESPACE}.whole_course_translations_enabled"
)

# .. toggle_name: PROBLEM_BLOCK_FEEDBACK_TRANSLATION
# .. toggle_implementation: CourseWaffleFlag
# .. toggle_default: False
# .. toggle_description: Enabling this feature allows ProblemBlock feedback to be translated
# .. toggle_warning: Requires ai-translations IDA to be available
# .. toggle_use_cases: opt_in
# .. toggle_creation_date: 2024-27-09
PROBLEM_BLOCK_FEEDBACK_TRANSLATION_FLAG = (
    f"{WAFFLE_NAMESPACE}.problem_block_feedback_translation"
)


class TranslatableXBlocksFeatureConfig:
    """Class containing configuration information about Translations features."""

    @staticmethod
    def xpert_translations_available_for_course(course_key):
        """
        Determine if the Xpert translation feature is available for the given context.

        Feature is considered "available" if we have the WHOLE_COURSE_TRANSLATIONS_FLAG
        enabled for that context.
        """
        xpert_translations_available_setting = course_waffle_flag(
            WHOLE_COURSE_TRANSLATIONS_FLAG, __name__, LOG_PREFIX
        )
        return xpert_translations_available_setting.is_enabled(course_key=course_key)

    @staticmethod
    def xpert_translations_enabled_for_course(course_key):
        """
        Determine if the Xpert translation feature is enabled for the given context.

        Feature is considered "enabled" if the following are both true:
        1) Feature is available
        2) Feature is enabled
        """
        xpert_translations_enabled_setting = course_waffle_flag(
            WHOLE_COURSE_TRANSLATIONS_ENABLED_FLAG, __name__, LOG_PREFIX
        )

        return TranslatableXBlocksFeatureConfig.xpert_translations_available_for_course(
            course_key
        ) and xpert_translations_enabled_setting.is_enabled(course_key=course_key)

    @staticmethod
    def problem_block_feedback_translation_enabled(course_key):
        """
        Determine if the ProblemBlock feedback translation feature is available for the given context.

        Feature is considered "enabled" if we have the PROBLEM_BLOCK_FEEDBACK_TRANSLATION_FLAG
        enabled for that context.
        """
        problem_block_feedback_translation = course_waffle_flag(
            PROBLEM_BLOCK_FEEDBACK_TRANSLATION_FLAG, __name__, LOG_PREFIX
        )
        return problem_block_feedback_translation.is_enabled(course_key=course_key)

    @staticmethod
    def enable_xpert_translations_for_course(course_key, enabled):
        """
        Set the waffle flag programatically for a course.

        Raises: ValueError if trying to modify "enabled" when feature not available.
        """
        # Shorthand to be less verbose
        config = TranslatableXBlocksFeatureConfig

        if not config.xpert_translations_available_for_course(course_key):
            err_string = (
                f"Cannot set {WHOLE_COURSE_TRANSLATIONS_ENABLED_FLAG}"
                f"when {WHOLE_COURSE_TRANSLATIONS_FLAG} is not enabled"
            )
            logger.error(err_string)
            raise ValueError(err_string)

        # Runtime import to allow tests to work
        WaffleFlagCourseOverrideModel = import_course_waffle_flag_override()

        WaffleFlagCourseOverrideModel.objects.create(
            waffle_flag=WHOLE_COURSE_TRANSLATIONS_ENABLED_FLAG,
            course_id=course_key,
            enabled=True,
            override_choice="on" if enabled else "off",
        )

        logger.info(
            f"Set {WHOLE_COURSE_TRANSLATIONS_ENABLED_FLAG} "
            f'to {"enabled" if enabled else "disabled"} '
            f"for course {str(course_key)}"
        )



"""
Filters required for added translation behavior.

See https://docs.openedx.org/projects/openedx-filters for docs.

NOTE that for the pipeline step to be enabled, you must add configuration to your environment's
OPEN_EDX_FILTERS_CONFIG as in translatable_xblocks/settings/devstack.py.
"""

# run_filter can have a number of different args, that is desired behavior
# pylint: disable=arguments-differ

from django.utils.translation import to_language
from openedx_filters.filters import PipelineStep


class UpdateRequestLanguageCode(PipelineStep):
    """
    Stop certificate creation if user is not in third party service.
    """

    def run_filter(self, context, student_view_context):
        """Filter step to update language code for requests which have requested a translation."""
        if student_view_context.get("src_lang") and student_view_context.get(
            "dest_lang"
        ):
            translation_language = student_view_context.get("dest_lang")
            context["LANGUAGE_CODE"] = to_language(translation_language)

        return context, student_view_context



#!/usr/bin/env python
"""
Package metadata for translatable_xblocks.
"""
import os
import re
import sys

from setuptools import find_packages, setup


def get_version(*file_paths):
    """
    Extract the version string from the file.

    Input:
     - file_paths: relative path fragments to file with
                   version string
    """
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename, encoding="utf8").read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def load_requirements(*requirements_paths):
    """
    Load all requirements from the specified requirements files.

    Requirements will include any constraints from files specified
    with -c in the requirements files.
    Returns a list of requirement strings.
    """
    # e.g. {"django": "Django", "confluent-kafka": "confluent_kafka[avro]"}
    by_canonical_name = {}

    def check_name_consistent(package):
        """
        Raise exception if package is named different ways.

        This ensures that packages are named consistently so we can match
        constraints to packages. It also ensures that if we require a package
        with extras we don't constrain it without mentioning the extras (since
        that too would interfere with matching constraints.)
        """
        canonical = package.lower().replace("_", "-").split("[")[0]
        seen_spelling = by_canonical_name.get(canonical)
        if seen_spelling is None:
            by_canonical_name[canonical] = package
        elif seen_spelling != package:
            raise Exception(
                f'Encountered both "{seen_spelling}" and "{package}" in requirements '
                "and constraints files; please use just one or the other."
            )

    requirements = {}
    constraint_files = set()

    # groups "pkg<=x.y.z,..." into ("pkg", "<=x.y.z,...")
    re_package_name_base_chars = r"a-zA-Z0-9\-_."  # chars allowed in base package name
    # Two groups: name[maybe,extras], and optionally a constraint
    requirement_line_regex = re.compile(
        r"([%s]+(?:\[[%s,\s]+\])?)([<>=][^#\s]+)?"
        % (re_package_name_base_chars, re_package_name_base_chars)
    )

    def add_version_constraint_or_raise(
        current_line, current_requirements, add_if_not_present
    ):
        regex_match = requirement_line_regex.match(current_line)
        if regex_match:
            package = regex_match.group(1)
            version_constraints = regex_match.group(2)
            check_name_consistent(package)
            existing_version_constraints = current_requirements.get(package, None)
            # It's fine to add constraints to an unconstrained package,
            # but raise an error if there are already constraints in place.
            if (
                existing_version_constraints
                and existing_version_constraints != version_constraints
            ):
                raise BaseException(
                    f"Multiple constraint definitions found for {package}:"
                    f' "{existing_version_constraints}" and "{version_constraints}".'
                    f"Combine constraints into one location with {package}"
                    f"{existing_version_constraints},{version_constraints}."
                )
            if add_if_not_present or package in current_requirements:
                current_requirements[package] = version_constraints

    # Read requirements from .in files and store the path to any
    # constraint files that are pulled in.
    for path in requirements_paths:
        with open(path) as reqs:
            for line in reqs:
                if is_requirement(line):
                    add_version_constraint_or_raise(line, requirements, True)
                if line and line.startswith("-c") and not line.startswith("-c http"):
                    constraint_files.add(
                        os.path.dirname(path)
                        + "/"
                        + line.split("#")[0].replace("-c", "").strip()
                    )

    # process constraint files: add constraints to existing requirements
    for constraint_file in constraint_files:
        with open(constraint_file) as reader:
            for line in reader:
                if is_requirement(line):
                    add_version_constraint_or_raise(line, requirements, False)

    # process back into list of pkg><=constraints strings
    constrained_requirements = [
        f'{pkg}{version or ""}' for (pkg, version) in sorted(requirements.items())
    ]
    return constrained_requirements


def is_requirement(line):
    """
    Return True if the requirement line is a package requirement.

    Returns:
        bool: True if the line is not blank, a comment,
        a URL, or an included file
    """
    return (
        line and line.strip() and not line.startswith(("-r", "#", "-e", "git+", "-c"))
    )


VERSION = get_version("translatable_xblocks", "__init__.py")

if sys.argv[-1] == "tag":
    print("Tagging the version on github:")
    os.system("git tag -a %s -m 'version %s'" % (VERSION, VERSION))
    os.system("git push --tags")
    sys.exit()

README = open(
    os.path.join(os.path.dirname(__file__), "README.rst"), encoding="utf8"
).read()
CHANGELOG = open(
    os.path.join(os.path.dirname(__file__), "CHANGELOG.rst"), encoding="utf8"
).read()

setup(
    name="translatable-xblocks",
    version=VERSION,
    description="""Plugin for adding translation behavior to XBlocks""",
    long_description=README + "\n\n" + CHANGELOG,
    author="Nathan Sprenkle",
    author_email="nsprenkle@2u.org",
    url="https://github.com/edx/translatable-xblocks",
    packages=find_packages(
        include=["translatable_xblocks", "translatable_xblocks.*"],
        exclude=["*tests", "tests", "test_utils"],
    ),
    entry_points={
        "lms.djangoapp": [
            "translatable_xblocks = translatable_xblocks.apps:TranslatableXBlocksConfig",
        ],
        "cms.djangoapp": [
            "translatable_xblocks = translatable_xblocks.apps:TranslatableXBlocksConfig",
        ],
        "xblock.v1.overrides": [
            "html = translatable_xblocks.xblocks.html_block:HtmlBlock",
            "problem = translatable_xblocks.xblocks.capa_block:ProblemBlock",
            "video = translatable_xblocks.xblocks.video_block:VideoBlock",
            "poll = translatable_xblocks.xblocks.poll_block:PollBlock",
        ],
    },
    include_package_data=True,
    install_requires=load_requirements("requirements/base.in"),
    tests_require=load_requirements("requirements/test.in"),
    python_requires=">=3.11",
    license="AGPL-3.0-or-later",
    zip_safe=False,
    keywords="Python edx",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Django",
        "Framework :: Django :: 4.2",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.11",
    ],
)



"""
Translatable version of Poll XBlock.
"""

# pylint:  disable=unnecessary-lambda-assignment

import importlib.resources
import inspect

from poll import PollBlock as OverriddenPollBlock
from xblock.core import XBlock
from xblock.fields import Scope

from translatable_xblocks.base import TranslatableXBlock
from translatable_xblocks.fields import TranslatableList, TranslatableString

# Make '_' a no-op so we can scrape strings. Using lambda instead of
#  `django.utils.translation.gettext_noop` because Django cannot be imported in this file
_ = lambda text: text


@XBlock.needs("i18n")
@XBlock.needs("settings")
@XBlock.needs("user")
class PollBlock(TranslatableXBlock, OverriddenPollBlock):
    """
    Translatable version of the PollBlock.
    """

    @classmethod
    def _open_resource(cls, uri):
        """HACK - override _open_resource method to load resource files from the original PollBlock."""
        return (
            importlib.resources.files(
                inspect.getmodule(OverriddenPollBlock).__package__
            )
            .joinpath(cls.resources_dir)
            .joinpath(uri)
            .open("rb")
        )

    display_name = TranslatableString(default=_("Poll"))
    question = TranslatableString(default=_("What is your favorite color?"))
    answers = TranslatableList(
        default=[
            ("R", {"label": _("Red"), "img": None, "img_alt": None}),
            ("B", {"label": _("Blue"), "img": None, "img_alt": None}),
            ("G", {"label": _("Green"), "img": None, "img_alt": None}),
            ("O", {"label": _("Other"), "img": None, "img_alt": None}),
        ],
        scope=Scope.settings,
        help=_("The answer options on this poll."),
    )

    @XBlock.supports("multi_device")
    def student_view(self, context, **kwargs):
        return super().student_view(context, **kwargs)
